#   ticket-db.pl - Bibliothek fuer das Infodrom Trouble Ticket System
#   Copyright (c) 1997,8  Martin Schulze <joey@infodrom.north.de>

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111, USA.

$ticket_version = "0.4.5.10";

# is_visible (groups,insertp)
#   Decides wether a ticket is visibe
#
# ticket_is_visible (nr)
#   Decides if the referenced ticket is visible, based on is_visible()
#
# has_access (nr, user)
#   Decides wether the referenced ticket is visible by that user.
#
# eval_tickets (dbh, where)
#   Queries the database and builds a binary tree, needed for viewing.
#   Modifies %ticket_tree and @tickets.
#
# is_second (nr)
#   Evaluates if the referred tickets is a child of some other.  Needs
#   %ticket_tree which is generated by eval_tickets()
#
# print_tickets (text, open)
#   Prints the whole received tickets, needs %ticket_tree
#
# print_table_ticket (nr, bold/$anchor, time)
#   Prints one ticket - and its children
#
# print_children (parent, text, open)
#   Prints the list of children
#
# has_open_children
#   Evaluates %ticket_tree and returns appropriate if open
#
# get_next_nr
#   Gibt die naechste freie Nr. in der Datenbank zurueck.  Wenn der
#   INSERT nicht schnell darauf folgt, kann es passieren, dass die
#   Nr. anschliessend bereits wieder vergeben ist.
#
# ticket_is_valid
#   Prueft, ob das angegebene Ticket gueltig ist.
#
# ticket_is_closed
#   Prueft, ob das Ticket bereits geschlossen ist oder nicht.
#
# ticket_insert
#   Prueft, ob das Ticket gueltig ist und fuegt es dann ein.  Die
#   INSERT-Mail wird automatisch verschickt.  Die Nummer des neuen
#   Tickets wird zurueckgegeben

require 'ticket.pl';

@ticket_priorities = ("high","medium","low","remind");

# [0] Gruppen des Tickets
# [1] insertp
# Needs @ticket_mygroups
sub is_visible
{
    my $groups = $_[0];
    my $insertp = $_[1];
    my $mygroup;
    my $mygroups;

    return 1 if ($ENV{'LOGNAME'} eq $ticket_superuser);
    return 1 if ($groups =~ /:Public:/);
    return 1 if ($insertp eq &user_email($ENV{'REMOTE_USER'}));

    if ($#ticket_mygroups < 0) {
	$mygroups = &user_groups($ENV{'REMOTE_USER'});
	$mygroups =~ s/^:(.*):$/$1/;
	@ticket_mygroups = split (/:/,$mygroups);
    }
    return 1 if (&user_groups($ENV{'REMOTE_USER'}) =~ /:All:/);

    foreach $mygroup (@ticket_mygroups) {
	return 1 if ($groups =~ /:$mygroup:/);
    }

    return 0;
}

# [0] = Nr.
sub ticket_is_visible
{
    my $nr = shift;
    my $query;
    my $sth;
    my $rc;
    my @row;

    return 1 if ($ENV{'LOGNAME'} eq $ticket_superuser);

    $query = sprintf("SELECT groups,insertp FROM $ticket_table WHERE nr = %d", $nr);
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0 && (@row = $sth->fetchrow_array)) {
	    return &is_visible($row[0], $row[1]);
	}
    }
    return 0;
}

# [0] Ticket #
# [1] User
sub has_access
{
    my $nr = shift;
    my $user = shift;
    my $query;
    my $sth;
    my $rc;
    my @row;
    my $i;

    $query = "SELECT insertp,changep";
    $i=0;while ($i < $ticket_staff) {
	$query .= sprintf (",staff%d", $i);
	$i++;
    }
    $query .= sprintf(" FROM $ticket_table WHERE nr = %d", $nr);
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0 && (@row = $sth->fetchrow_array)) {
	    return 1 if (&email_user($row[0]) eq $user);
	    return 1 if (&email_user($row[1]) eq $user);
	    $i=0;while ($i < $ticket_staff) {
		return 1 if ($row[2+$i] eq $user);
		$i++;
	    }
	}
    }
    return 0;
}

@tickets = ();
%ticket_tree = ();
$ticket_spacing = "";
sub eval_tickets
{
    my $dbh = shift;
    my $where = shift;
    my $query;
    my $sth;
    my $rc;
    my @row;
    my $children;
    my $nr;
    my $check;

    my $is_admin = undef;

    $is_admin = 1 if (defined $cgi && defined $cgi->param('admin') && &is_admin($ENV{'REMOTE_USER'}));

    $query = "SELECT nr,anchor,priority,staff0,groups,insertp FROM $ticket_table $where ORDER BY priority,staff0,nr";
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0 ) {
	    while (@row = $sth->fetchrow_array) {
		if (!$is_admin) {
		    next unless (&is_visible($row[4], $row[5]));
		}
		push (@tickets, $row[0]);
		if ($row[1]) {
		    $children = $ticket_tree{$row[1]};
		    $children = ":" if (!length($children)) ;
		    $children .= $row[0] . ":";
		    $ticket_tree{$row[1]} = $children;
		    $children = ":";
		} else {
		    $ticket_tree{$row[0]} = ":" if (!$ticket_tree{$row[0]});
		}
	    }

	    # Sanity check
	    foreach $check (keys(%ticket_tree)) {
		if (! grep (/^($check)$/, @tickets)) {
		    $children = $ticket_tree{$check};
		    $children =~ s/^:(.*):$/$1/;
		    if (length($children)) {
			@row = split(/:/,$children);
			foreach $nr (@row) {
			    $ticket_tree{$nr} = ":" if (!exists ($ticket_tree{$nr}));
			}
			delete $ticket_tree{$check};
		    }
		}
	    }
	}
    }
}

# [0] Nr
#   Evaluates if the referred tickets is a child of some other.  Needs
#   %ticket_tree which is generated by eval_tickets()
#
sub is_second
{
    return 0 if ($#_ < 0);
    foreach $key (keys(%ticket_tree)) {
	return 1 if ($ticket_tree{$key} =~ /:$_[0]:/);
    }
    return 0;
}

# [0] Nr
#   Evaluates if the referred tickets has one or more children.
#   Needs %ticket_tree which is generated by eval_tickets()
#
sub has_children
{
    return 0 if ($#_ < 0);
    foreach $key (keys(%ticket_tree)) {
	return 1 if ($ticket_tree{$key} =~ /^$_[0]:/);
    }
    return 0;
}

sub print_tickets
{
    my $text = shift;
    my $open = shift;
    my $urladd = shift;
    my $count = 0;
    my $do_table = ($ticket_browser_table && $ticket_intern{'misc'} =~ /tablelist/);

    $ticket_spacing = "";
    if (!$text) {
	if ($do_table) {
	    print "<table border=1 cellspacing=0 cellpadding=2>\n";
	    print "<tr bgcolor=\"#c0c0c0\">\n";
	    print "  <th width=30>Nr.</th>\n";
	    print "  <th width=30>Pri</th>\n";
	    print "  <th width=100>Staff</th>\n";
	    print "  <th width=70%>Subject</th>\n";
	    if ($open) {
		print "  <th>Time</th>\n";
	    } else {
		print "  <th>Modified</th>\n";
	    }
	    print "</tr>\n";
	} else {
	    print "<ul>\n";
	}
    }

    foreach $nr (@tickets) {
	if (! &is_second ($nr)) {
	    do print_table_ticket ($nr, "bold", $text, $open, $urladd, $do_table);
	    $count++;
	    $count += &print_children($nr, $text, $open, $urladd, $do_table);
	}
    }
    if (!$text) {
	if ($do_table) {
	    print  "<tr>\n";
	    printf "  <th colspan=5 align=left>%d Tickets</th>\n", $count;
	    print  "</tr>\n";
	    print  "</table>\n";
	} else {
	    print "</ul>\n";
	    printf "<b>%d Tickets</b><p>\n\n", $count;
	}
    } else {
	printf "\n%d Tickets\n\n", $count;
    }
}

# [0] Nr.
# [1] bold/$anchor
# [2] 1/0 (=time ja/nein)
# [3] Appendix to URL like "&admin=y"
#
sub print_table_ticket
{
    my $nr = shift;
    my $bold = shift;
    my $text = shift;
    my $open = shift;
    my $urladd = shift;
    my $do_table = shift;
    my $query;
    my $sth;
    my $rc;
    my @row;
    my $td;
    my $subject;
    my $state;

    if ($bold eq "bold") {
	$td = "th";
	$anchor = 0;
    } else {
	$anchor = $bold;
	$td = "td";
    }

    $query  = "SELECT priority,staff0,subject";
    if ($typ == 1) {
	$query .= ",closetime";
    } else {
	$query .= ",changed";
	$query .= ",closing" if ($open == 2);
    }
    $query .= " FROM $ticket_table WHERE nr = $nr";
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0 && (@row = $sth->fetchrow_array)) {
	    if ($text) {
		if ($open == 1) {
		    printf "%s#%d %s [%s, %s, %d min]\n",
		    $ticket_spacing, $nr, $row[2], $ticket_priorities[$row[0]], $row[1], $row[3];
		} elsif ($open == 2) {
		    if (length($row[4])) { $state = "closed"; }
		    else { $state = "open"; }
		    printf "%s#%d %s [%s, %s, %s %s]\n",
		    $ticket_spacing, $nr, $row[2], $ticket_priorities[$row[0]],
		        $row[1], &string_to_date($row[3]), $state;
		} else {
		    printf "%s#%d %s [%s, %s, %s]\n",
		    $ticket_spacing, $nr, $row[2], $ticket_priorities[$row[0]], $row[1], &string_to_date($row[3]);
		}
	    } else {
	    if ($do_table) {
		print "<tr>\n";
		printf "  <td align=right>%d</td>\n", $nr;
		printf "  <td align=center>%s</td>\n", ('HI','M','LO')[$row[0]];
		printf "  <td align=center>%s</td>\n", $row[1];
		$subject = $row[2];
		$subject = sprintf ("(%d) %s", $anchor, $subject) if ($anchor);
		printf "  <$td align=left><a href=\"%s/listone?nr=%d%s\">%s</a></$td>\n",
		    $ticket_html_base, $nr, $urladd, $subject;
		if ($open == 1) {
		    printf "  <td align=right>%d</td>\n", $row[3];
		} elsif ($typ == 2) {
		    if (length($row[4])) { $state = "-"; }
		    else { $state = "+"; }
		    printf "  <td align=center>%s %s</td>\n", &string_to_date($row[3]), $state;
		} else {
		    printf "  <td align=center>%s</td>\n", &string_to_date($row[3])
		}
#		printf "  <td>%s by %s</td>\n", &string_to_date($row[3]), &email_user($row[4]);
		print "</tr>\n";
	    } else {
		if ($open == 1) {
		    printf "<li> <a href=\"%s/listone?nr=%d%s\">%s</a> [%s, %s, %d min]\n",
		    $ticket_html_base, $nr, $urladd, $row[2], $ticket_priorities[$row[0]], $row[1], $row[3];
		} elsif ($open == 2) {
		    if (length($row[4])) { $state = "closed"; }
		    else { $state = "open"; }
		    printf "<li> <a href=\"%s/listone?nr=%d%s\">%s</a> [%s, %s, %s, %s]\n",
		    $ticket_html_base, $nr, $urladd, $row[2], $ticket_priorities[$row[0]], $row[1],
		    &string_to_date($row[3]), $state;
		} else {
		    printf "<li> <a href=\"%s/listone?nr=%d%s\">%s</a> [%s, %s, %s]\n",
		    $ticket_html_base, $nr, $urladd, $row[2], $ticket_priorities[$row[0]], $row[1], &string_to_date($row[3]);
		}
	    }
	}

	}
    } else {
	do error_query($query);
    }
}

sub print_children
{
    my $parent = shift;
    my $text = shift;
    my $open = shift;
    my $urladd = shift;
    my $do_table = shift;
    my @list;
    my $second;
    my $count = 0;

    $second = $ticket_tree{$parent};
    $second =~ s/^:(.*):$/$1/;
    if (length($second)) {
	@list = split(/:/,$second);
	if (@list) {
	    print "<ul>\n" if (!$text && !$do_table);
	foreach $nr (@list) {
	    $ticket_spacing .= "  ";
	    do print_table_ticket ($nr, $parent, $text, $open, $urladd, $do_table);
	    $count++;
	    $count += &print_children ($nr,$text,$open,$urladd,$do_table);
	    chop($ticket_spacing);
	    chop($ticket_spacing);
	}
	print "</ul>\n" if (!$text && !$do_table);
	}
    }
    return $count;
}

sub has_open_children
{
    my $nr = shift;
    my $open = shift;
    my $query;
    my $sth;
    my $rc;
    my @row;

    return 0 if (!$nr);

    $query = "SELECT nr FROM $ticket_table WHERE anchor = $nr AND closing = '' ORDER BY nr";
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0) {
	    return 1;
	}
    }
    return 0;
}

########### get_class_descrip
#
# Gibt die Beschreibung einer Klasse zurueck
#
sub get_class_descrip
{
    my $class = shift;
    my $query;
    my $sth;
    my $rc;
    my @row;

    return if (!$class);

    $query = "SELECT description FROM $ticket_table_classes WHERE name = '$class'";
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0 && (@row = $sth->fetchrow_array)) {
	    return ($row[0]);
	}
    } else {
	do error_query($query);
    }
}

########### get_next_nr
#
# Gibt die naechste freie Nr. in der Datenbank zurueck.  Wenn der
# INSERT nicht schnell darauf folgt, kann es passieren, dass die
# Nr. anschliessend bereits wieder vergeben ist.
#
sub get_next_nr
{
    my $query;
    my $sth;
    my $rc;
    my @row;

    $dbh = &noc_connect if (!$dbh);

    $query = "SELECT nr FROM $ticket_table ORDER BY nr DESC";
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0 && (@row = $sth->fetchrow_array)) {
	    return $row[0] + 1;
	}
    }
    return 1;
}

########### ticket_is_valid
#
# Prueft, ob das angegebene Ticket gueltig ist.
#
# Args: "text/html", length(subject), @staff, groups, length(text)
#       groups ist mit Leerzeichen getrennt
#
sub ticket_is_valid
{
    my $type = shift;
    my $lsubject = shift;
    my $staff = shift;
    my @staff = @$staff;
    my $groups = shift;
    my $lbody = shift;
    my $res = 1;
    my @groups;
    my $group;
    my $tickgroups;
    my $ok;
    my $i;

    if (!$lsubject) {
	print "<h3>" . sprintf (&gettext("The field '%s' must not be empty!"), "Subject") . "</h3>\n\n";
	$res = 0;
    } elsif ($lsubject>$ticket_length{'subject'}-1) {
	print "<h3>" . sprintf (&gettext("The field '%s' must not contain more than %d characters!"), "Subject", $ticket_length{'subject'}) . "</h3>\n\n";
	$res = 0;
    }
    if ($#staff == -1) {
	print "<h3>" . sprintf (&gettext("The field '%s' must not be empty!"), &gettext("Staff")) . "</h3>\n\n";
	$res = 0;
    }
    if (!$lbody) {
	print "<h3>" . sprintf (&gettext("The field '%s' must not be empty!"), &gettext("Text")) . "</h3>\n\n";
	$res = 0;
    } elsif ($lsubject>$ticket_length{'subject'}-1) {
	print "<h3>" . sprintf (&gettext("The field '%s' must not contain more than %d characters!"), &gettext("Text"), $ticket_length{'body'}) . "</h3>\n\n";
	$res = 0;
    }
    if ($staff[0] ne $ENV{'REMOTE_USER'}) {
	if (!length($groups)) {
	    print "<h3>" . sprintf (&gettext("The field '%s' must not be empty!"), &gettext("Groups")) . "</h3>\n\n";
	    $res = 0;
	}
    }

    if (length($groups)) {
	@groups = split (/ /,$groups);
	$tickgroups = ":" . join (":", @groups) . ":";
    }
    if ($tickgroups !~ /:Public:/) {

	# Staff0
	$i = 0;
	while ($i < $ticket_staff) {
	    if ($staff[$i] && ($staff[$i] ne $ENV{'REMOTE_USER'})) {
		$groups = &user_groups($staff[$i]);
		$groups =~ s/^:(.*):$/$1/;
		@groups = split(/:/,$groups);
		$ok = 0;
		foreach $group (@groups) {
		    $ok = 1 if ($tickgroups =~ /:$group:/);
		}
		if (!$ok) {
		    print "<h3>" . &gettext("Not enough groups specified!") . "</h3>\n\n";
		    print sprintf (&gettext("%s would not be able to access the ticket."), &user_realname($staff[0])) . "<p>\n\n";
		    $res = 0;
		}
	    }
	    $i++;
	}
    }
    return $res;
}

########### ticket_is_closed
#
# Prueft, ob das angegebene Ticket bereits geschlossen ist
#
# Args: Nummer
#
sub ticket_is_closed
{
    my $nr = shift;
    my $query;
    my $sth;
    my $rc;
    my @row;

    $dbh = &noc_connect if (!$dbh);

    $query = "SELECT closing FROM $ticket_table WHERE nr = $nr";
    $sth = $dbh->prepare($query);
    if ($sth) {
	$rc = $sth->execute;
	if ($rc > 0 && (@row = $sth->fetchrow_array)) {
	    return 1 if (length($row[0]));
	}
    }
    return 0;
}

########### ticket_insert
#
# Prueft, ob das Ticket gueltig ist und fuegt es dann ein.  Die
# INSERT-Mail wird automatisch verschickt.  Die Nummer des neuen
# Tickets wird zurueckgegeben
#
# Args: (user,subject,class,priority,@staff,groups,anchor,bodypre,body,fixdate)
#       groups ist mit Leerzeichen getrennt
#
sub ticket_insert
{
    my ($nr) = 0;
    my $user = shift;
    my $subject = shift;
    my $class = shift;
    my $priority = shift;
    my $staff = shift;
    my @staff = @$staff;
    my $groups = shift;
    my $anchor = shift;
    my $bodypre = shift;
    my $body = shift;
    my $fixdate = shift;
    my $mygroups = $groups;
    my $query, $sth, $rc;
    my $datastring;
    my $i;

    $ENV{'REMOTE_USER'} = $user if (! scalar $ENV{'REMOTE_USER'});

    $priority = 3 if ($priority < 0 || $priority > 3);
    # return 0 if (!is_class($class));
    # return 0 if (!ticket_exists($anhor));

    $dbh = &noc_connect if (!$dbh);

    $subject =~ s/\s*$//;
    if (&ticket_is_valid("text",length($subject), \@staff, $groups, length($body)) && $dbh) {
	$date = &stringdate();
	$mygroups =~ s/ /:/g;
	$mygroups = ":" . $mygroups . ":";

	$datastring = sprintf ("%s,'%s',%d,",
			       $dbh->quote(substr($subject,0,$ticket_length{'subject'}-1)),
			       $class, $priority);
	$i=0;
	while ($i < $ticket_staff) {
	    $datastring .= sprintf("'%s',", $staff[$i]);
	    $i++;
	}
	$datastring .= sprintf("'%s',%d,'%s',%s,'%s','%s','%s','','',0,'n','','%s','%s')",
			  $mygroups,$anchor, $bodypre,
			  $dbh->quote(substr(&prepare_text($body),0,
				      $ticket_length{'body'}-1)),
			  &date_to_string($fixdate),
			  $date, &user_email($ENV{'REMOTE_USER'}),
			  $date, &user_email($ENV{'REMOTE_USER'}));
	do {
	    $nr = &get_next_nr();
	    $query = sprintf ("INSERT INTO tickets VALUES (%d,%s", $nr, $datastring);
	    $rc = $dbh->do($query);
	    $rc = -1 if (!$rc && $dbh->errstr =~ /non unique/i);
	    if (!$rc) {
		printf STDERR "Query: %s\n", $query;
		printf STDERR "Error: %s\n", $dbh->errstr;
	    }
	} until ($rc);

	if ($rc) {
	    open (MAIL, "|/usr/lib/sendmail -t");
	    push (@staff, $user);
	    printf MAIL "To: %s\n", join(", ", &uniq_user_email(@staff));
	    pop (@staff);
	    printf MAIL "From: %s\n", $ticket_from;
	    printf MAIL "Bcc: %s\n", $ticket_bcc if ($ticket_bcc);
	    printf MAIL "Subject: [INSERT] #%d: %s\n", $nr, &de_html($subject);
	    print  MAIL "Precedence: junk\n";
	    printf MAIL "X-Mailer: ticket %s\n", $ticket_version;
	    print  MAIL "\n";
	    printf MAIL &gettext("%s has opened the following ticket on %s:") . "\n\n",
		&user_realname($ENV{'REMOTE_USER'}), &string_to_date(&stringdate());
	    printf MAIL "Ticket ....: #%d\n", $nr;
	    printf MAIL "Subject ...: %s\n", &de_html($subject);
	    printf MAIL "Klasse ....: %s\n", &get_class_descrip($class);
	    printf MAIL "Prioritaet : %s\n", $ticket_priorities[$priority];
	    printf MAIL "Mitarbeiter: %s\n", join(", ", &uniq_user_realname(@staff));
	    printf MAIL "Gruppen ...: %s\n", join(", ", split(/:/, $groups)) if (length($groups));
	    printf MAIL "Auftrag ...: #%d (%s)\n", $anchor, &get_subject($anchor) 
		if ($anchor);
	    if (exists $ENV{'SERVER_NAME'} && exists $ENV{'SERVER_PORT'}) {
		$url = "http://" . $ENV{'SERVER_NAME'};
		$url .= ":" . $ENV{'SERVER_PORT'} if ($ENV{'SERVER_PORT'} != 80);
		$url .= $ticket_html_base . "/listone?nr=" . $nr;
		printf MAIL "URL .......: %s\n", $url;
	    }
	    printf MAIL "\n%s\n\n", &de_html($body);
	    print  MAIL sprintf (&gettext("%s was set as fixdate."), $fixdate) . "\n\n" if (length($fixdate));
	    printf MAIL "\n%s,\n\n\t%s\n", &gettext("Best regards"), &gettext("Your Ticket System");
	    print  MAIL $ticket_signature;
	    close (MAIL);
	} else {
	    $nr = 0;
	}
    } else {
	printf "<h3>%s</h3>\n\n", &gettext("The ticket is invalid.");
    }

    return $nr
}

# Make perl happy
#
1;
